@@ -3,11 +3,13 @@
 import sys
 
 import click
+from mysql.connector import CharacterSet
 from tabulate import tabulate
 
 from . import SQLite3toMySQL
 from .click_utils import OptionEatAll, prompt_password
 from .debug_info import info
+from .mysql_utils import mysql_supported_character_sets
 
 
 @click.command()
@@ -62,6 +64,24 @@ from .debug_info import info
     default="VARCHAR(255)",
     help="MySQL default string field type. Defaults to VARCHAR(255).",
 )
+@click.option(
+    "--mysql-charset",
+    metavar="TEXT",
+    type=click.Choice(list(CharacterSet.get_supported()), case_sensitive=False),
+    default="utf8mb4",
+    show_default=True,
+    help="MySQL database and table character set",
+)
+@click.option(
+    "--mysql-collation",
+    metavar="TEXT",
+    type=click.Choice(
+        [charset.collation for charset in mysql_supported_character_sets()],
+        case_sensitive=False,
+    ),
+    default=None,
+    help="MySQL database and table collation",
+)
 @click.option(
     "-T",
     "--use-fulltext",
@@ -91,6 +111,8 @@ def cli(
     skip_ssl,
     mysql_integer_type,
     mysql_string_type,
+    mysql_charset,
+    mysql_collation,
     use_fulltext,
     with_rowid,
     chunk,
@@ -99,6 +121,21 @@ def cli(
 ):
     """Transfer SQLite to MySQL using the provided CLI options."""
     try:
+        if mysql_collation:
+            charset_collations = tuple(
+                cs.collation
+                for cs in mysql_supported_character_sets(mysql_charset.lower())
+            )
+            if mysql_collation not in set(charset_collations):
+                raise click.ClickException(
+                    "Error: Invalid value for '--collation' of charset '{charset}': '{collation}' is not one of "
+                    "{collations}.".format(
+                        collation=mysql_collation,
+                        charset=mysql_charset,
+                        collations="'" + "', '".join(charset_collations) + "'",
+                    )
+                )
+
         converter = SQLite3toMySQL(
             sqlite_file=sqlite_file,
             sqlite_tables=sqlite_tables,
@@ -112,6 +149,8 @@ def cli(
             mysql_ssl_disabled=skip_ssl,
             mysql_integer_type=mysql_integer_type,
             mysql_string_type=mysql_string_type,
+            mysql_charset=mysql_charset.lower() if mysql_charset else "utf8mb4",
+            mysql_collation=mysql_collation.lower() if mysql_collation else None,
             use_fulltext=use_fulltext,
             with_rowid=with_rowid,
             chunk=chunk,
