@@ -13,8 +13,7 @@ from sys import stdout
 
 import mysql.connector
 import six
-from mysql.connector import errorcode
-from packaging import version
+from mysql.connector import CharacterSet, errorcode
 from tqdm import tqdm, trange
 
 from sqlite3_to_mysql.sqlite_utils import (
@@ -23,6 +22,14 @@ from sqlite3_to_mysql.sqlite_utils import (
     convert_blob,
     convert_decimal,
     convert_timedelta,
+    unicase_compare,
+)
+
+from .mysql_utils import (
+    MYSQL_COLUMN_TYPES,
+    check_mysql_fulltext_support,
+    check_mysql_json_support,
+    safe_identifier_length,
 )
 
 if six.PY2:
@@ -32,46 +39,6 @@ if six.PY2:
 class SQLite3toMySQL:
     """Use this class to transfer an SQLite 3 database to MySQL."""
 
-    # shamelessly copied from SQLAlchemy's dialects/mysql/__init__.py
-    MYSQL_COLUMN_TYPES = {
-        "BIGINT",
-        "BINARY",
-        "BIT",
-        "BLOB",
-        "BOOLEAN",
-        "CHAR",
-        "DATE",
-        "DATETIME",
-        "DECIMAL",
-        "DOUBLE",
-        "ENUM",
-        "DECIMAL",
-        "FLOAT",
-        "INTEGER",
-        "INTEGER",
-        "JSON",
-        "LONGBLOB",
-        "LONGTEXT",
-        "MEDIUMBLOB",
-        "MEDIUMINT",
-        "MEDIUMTEXT",
-        "NCHAR",
-        "NVARCHAR",
-        "NUMERIC",
-        "SET",
-        "SMALLINT",
-        "REAL",
-        "TEXT",
-        "TIME",
-        "TIMESTAMP",
-        "TINYBLOB",
-        "TINYINT",
-        "TINYTEXT",
-        "VARBINARY",
-        "VARCHAR",
-        "YEAR",
-    }
-
     COLUMN_PATTERN = re.compile(r"^[^(]+")
     COLUMN_LENGTH_PATTERN = re.compile(r"\(\d+\)$")
 
@@ -130,6 +97,18 @@ class SQLite3toMySQL:
             kwargs.get("mysql_string_type") or "VARCHAR(255)"
         ).upper()
 
+        self._mysql_charset = kwargs.get("mysql_charset") or "utf8mb4"
+
+        self._mysql_collation = (
+            kwargs.get("mysql_collation")
+            or CharacterSet.get_default_collation(self._mysql_charset.lower())[0]
+        )
+        if (
+            not kwargs.get("mysql_collation")
+            and self._mysql_collation == "utf8mb4_0900_ai_ci"
+        ):
+            self._mysql_collation = "utf8mb4_general_ci"
+
         self._use_fulltext = kwargs.get("use_fulltext") or False
 
         self._with_rowid = kwargs.get("with_rowid") or False
@@ -146,6 +125,7 @@ class SQLite3toMySQL:
             realpath(self._sqlite_file), detect_types=sqlite3.PARSE_DECLTYPES
         )
         self._sqlite.row_factory = sqlite3.Row
+        self._sqlite.create_collation("unicase", unicase_compare)
 
         self._sqlite_cur = self._sqlite.cursor()
 
@@ -173,10 +153,8 @@ class SQLite3toMySQL:
                     raise
 
             self._mysql_version = self._get_mysql_version()
-            self._mysql_json_support = self._check_mysql_json_support(
-                self._mysql_version
-            )
-            self._mysql_fulltext_support = self._check_mysql_fulltext_support(
+            self._mysql_json_support = check_mysql_json_support(self._mysql_version)
+            self._mysql_fulltext_support = check_mysql_fulltext_support(
                 self._mysql_version
             )
 
@@ -219,42 +197,6 @@ class SQLite3toMySQL:
             )
             raise
 
-    @staticmethod
-    def _check_mysql_json_support(version_string):
-        mysql_version = version.parse(re.sub("-.*$", "", version_string))
-
-        if version_string.lower().endswith("-mariadb"):
-            if (
-                mysql_version.major >= 10
-                and mysql_version.minor >= 2
-                and mysql_version.micro >= 7
-            ):
-                return True
-        else:
-            if mysql_version.major >= 8:
-                return True
-            if mysql_version.minor >= 7 and mysql_version.micro >= 8:
-                return True
-        return False
-
-    @staticmethod
-    def _check_mysql_fulltext_support(version_string):
-        mysql_version = version.parse(re.sub("-.*$", "", version_string))
-
-        if version_string.lower().endswith("-mariadb"):
-            if (
-                mysql_version.major >= 10
-                and mysql_version.minor >= 0
-                and mysql_version.micro >= 5
-            ):
-                return True
-        else:
-            if mysql_version.major >= 8:
-                return True
-            if mysql_version.minor >= 6:
-                return True
-        return False
-
     def _sqlite_table_has_rowid(self, table):
         try:
             self._sqlite_cur.execute("""SELECT rowid FROM "{}" LIMIT 1""".format(table))
@@ -267,11 +209,13 @@ class SQLite3toMySQL:
         try:
             self._mysql_cur.execute(
                 """
-                CREATE DATABASE IF NOT EXISTS `{}`
-                DEFAULT CHARACTER SET utf8mb4
-                DEFAULT COLLATE utf8mb4_general_ci
+                CREATE DATABASE IF NOT EXISTS `{database}`
+                DEFAULT CHARACTER SET {charset}
+                DEFAULT COLLATE {collation}
             """.format(
-                    self._mysql_database
+                    database=self._mysql_database,
+                    charset=self._mysql_charset,
+                    collation=self._mysql_collation,
                 )
             )
             self._mysql_cur.close()
@@ -326,7 +270,7 @@ class SQLite3toMySQL:
             return match.group(0).upper() + length
         if data_type in {"INT64", "NUMERIC"}:
             return "BIGINT" + self._column_type_length(column_type, 19)
-        if data_type not in self.MYSQL_COLUMN_TYPES:
+        if data_type not in MYSQL_COLUMN_TYPES:
             return self._mysql_string_type
         return full_column_type
 
@@ -342,7 +286,9 @@ class SQLite3toMySQL:
     def _create_table(self, table_name, transfer_rowid=False):
         primary_keys = []
 
-        sql = "CREATE TABLE IF NOT EXISTS `{}` ( ".format(table_name)
+        sql = "CREATE TABLE IF NOT EXISTS `{}` ( ".format(
+            safe_identifier_length(table_name)
+        )
 
         if transfer_rowid:
             sql += " `rowid` BIGINT NOT NULL, "
@@ -356,7 +302,7 @@ class SQLite3toMySQL:
         for row in rows:
             column = dict(row)
             sql += " `{name}` {type} {notnull} {auto_increment}, ".format(
-                name=column["name"],
+                name=safe_identifier_length(column["name"]),
                 type=self._translate_type_from_sqlite_to_mysql(column["type"]),
                 notnull="NOT NULL" if column["notnull"] or column["pk"] else "NULL",
                 auto_increment="AUTO_INCREMENT"
@@ -368,7 +314,10 @@ class SQLite3toMySQL:
                 else "",
             )
             if column["pk"] > 0:
-                primary_key = {"column": column["name"], "length": ""}
+                primary_key = {
+                    "column": safe_identifier_length(column["name"]),
+                    "length": "",
+                }
                 # In case we have a non-numeric primary key
                 column_type = self._translate_type_from_sqlite_to_mysql(column["type"])
                 if column_type in {"TEXT", "BLOB"} or column_type.startswith(
@@ -386,14 +335,23 @@ class SQLite3toMySQL:
                 )
             )
         if transfer_rowid:
-            sql += ", CONSTRAINT `{}_rowid` UNIQUE (`rowid`)".format(table_name)
-        sql += " ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_general_ci"
+            sql += ", CONSTRAINT `{}_rowid` UNIQUE (`rowid`)".format(
+                safe_identifier_length(table_name)
+            )
+        sql += " ) ENGINE=InnoDB DEFAULT CHARSET={charset} COLLATE={collation}".format(
+            charset=self._mysql_charset,
+            collation=self._mysql_collation,
+        )
 
         try:
             self._mysql_cur.execute(sql)
             self._mysql.commit()
         except mysql.connector.Error as err:
-            self._logger.error("MySQL failed creating table %s: %s", table_name, err)
+            self._logger.error(
+                "MySQL failed creating table %s: %s",
+                safe_identifier_length(table_name),
+                err,
+            )
             raise
 
     def _add_indices(self, table_name):
@@ -419,18 +377,21 @@ class SQLite3toMySQL:
                 table_columns[index_info["name"]].upper() == "TEXT"
                 for index_info in index_infos
             ):
-
                 if self._use_fulltext and self._mysql_fulltext_support:
                     # Use fulltext if requested and available
                     index_type = "FULLTEXT"
                     index_columns = ",".join(
-                        "`{}`".format(index_info["name"]) for index_info in index_infos
+                        "`{}`".format(safe_identifier_length(index_info["name"]))
+                        for index_info in index_infos
                     )
                 else:
                     # Limit the max TEXT field index length to 255
                     index_columns = ", ".join(
                         "`{column}`{length}".format(
-                            column=index_info["name"], length="({})".format(255)
+                            column=safe_identifier_length(index_info["name"]),
+                            length="({})".format(255)
+                            if table_columns[index_info["name"]].upper() == "TEXT"
+                            else "",
                         )
                         for index_info in index_infos
                     )
@@ -449,18 +410,40 @@ class SQLite3toMySQL:
                             index_length = suffix.group(0)
                     column_list.append(
                         "`{column}`{length}".format(
-                            column=index_info["name"], length=index_length
+                            column=safe_identifier_length(index_info["name"]),
+                            length=index_length,
                         )
                     )
                 index_columns = ", ".join(column_list)
 
-            self._add_index(
-                table_name=table_name,
-                index_type=index_type,
-                index=index,
-                index_columns=index_columns,
-                index_infos=index_infos,
-            )
+            try:
+                self._add_index(
+                    table_name=table_name,
+                    index_type=index_type,
+                    index=index,
+                    index_columns=index_columns,
+                    index_infos=index_infos,
+                )
+            except mysql.connector.Error as err:
+                if err.errno == errorcode.ER_BAD_FT_COLUMN and index_type == "FULLTEXT":
+                    # handle bad FULLTEXT index
+                    self._add_index(
+                        table_name=table_name,
+                        index_type="UNIQUE" if int(index["unique"]) == 1 else "INDEX",
+                        index=index,
+                        index_columns=", ".join(
+                            "`{column}`{length}".format(
+                                column=safe_identifier_length(index_info["name"]),
+                                length="({})".format(255)
+                                if table_columns[index_info["name"]].upper() == "TEXT"
+                                else "",
+                            )
+                            for index_info in index_infos
+                        ),
+                        index_infos=index_infos,
+                    )
+                else:
+                    raise
 
     def _add_index(
         self,
@@ -475,12 +458,13 @@ class SQLite3toMySQL:
             ALTER TABLE `{table}`
             ADD {index_type} `{name}`({columns})
         """.format(
-            table=table_name,
+            table=safe_identifier_length(table_name),
             index_type=index_type,
-            name=index["name"][:64]
+            name=safe_identifier_length(index["name"])
             if index_iteration == 0
             else "{name}_{index_iteration}".format(
-                name=index["name"][:60], index_iteration=index_iteration
+                name=safe_identifier_length(index["name"], max_length=60),
+                index_iteration=index_iteration,
             ),
             columns=index_columns,
         )
@@ -489,8 +473,11 @@ class SQLite3toMySQL:
             self._logger.info(
                 """Adding %s to column "%s" in table %s""",
                 "unique index" if int(index["unique"]) == 1 else "index",
-                ", ".join(index_info["name"] for index_info in index_infos),
-                table_name,
+                ", ".join(
+                    safe_identifier_length(index_info["name"])
+                    for index_info in index_infos
+                ),
+                safe_identifier_length(table_name),
             )
             self._mysql_cur.execute(sql)
             self._mysql.commit()
@@ -505,10 +492,25 @@ class SQLite3toMySQL:
                     index_infos=index_infos,
                     index_iteration=index_iteration + 1,
                 )
+            elif err.errno == errorcode.ER_BAD_FT_COLUMN:
+                # handle bad FULLTEXT index
+                self._logger.warning(
+                    """Failed adding FULLTEXT index to column "%s" in table %s. Retrying without FULLTEXT ...""",
+                    ", ".join(
+                        safe_identifier_length(index_info["name"])
+                        for index_info in index_infos
+                    ),
+                    safe_identifier_length(table_name),
+                )
+                raise
             else:
                 self._logger.error(
-                    "MySQL failed adding indices to table %s: %s",
-                    table_name,
+                    """MySQL failed adding index to column "%s" in table %s: %s""",
+                    ", ".join(
+                        safe_identifier_length(index_info["name"])
+                        for index_info in index_infos
+                    ),
+                    safe_identifier_length(table_name),
                     err,
                 )
                 raise
@@ -528,10 +530,10 @@ class SQLite3toMySQL:
             """.format(
                 id=foreign_key["id"],
                 seq=foreign_key["seq"],
-                table=table_name,
-                column=foreign_key["from"],
-                ref_table=foreign_key["table"],
-                ref_column=foreign_key["to"],
+                table=safe_identifier_length(table_name),
+                column=safe_identifier_length(foreign_key["from"]),
+                ref_table=safe_identifier_length(foreign_key["table"]),
+                ref_column=safe_identifier_length(foreign_key["to"]),
                 on_delete=foreign_key["on_delete"].upper()
                 if foreign_key["on_delete"].upper() != "SET DEFAULT"
                 else "NO ACTION",
@@ -543,17 +545,20 @@ class SQLite3toMySQL:
             try:
                 self._logger.info(
                     "Adding foreign key to %s.%s referencing %s.%s",
-                    table_name,
-                    foreign_key["from"],
-                    foreign_key["table"],
-                    foreign_key["to"],
+                    safe_identifier_length(table_name),
+                    safe_identifier_length(foreign_key["from"]),
+                    safe_identifier_length(foreign_key["table"]),
+                    safe_identifier_length(foreign_key["to"]),
                 )
                 self._mysql_cur.execute(sql)
                 self._mysql.commit()
             except mysql.connector.Error as err:
                 self._logger.error(
-                    "MySQL failed adding foreign keys to table %s: %s",
-                    table_name,
+                    "MySQL failed adding foreign key to %s.%s referencing %s.%s: %s",
+                    safe_identifier_length(table_name),
+                    safe_identifier_length(foreign_key["from"]),
+                    safe_identifier_length(foreign_key["table"]),
+                    safe_identifier_length(foreign_key["to"]),
                     err,
                 )
                 raise
@@ -640,13 +645,16 @@ class SQLite3toMySQL:
                             table_name=table["name"],
                         )
                     )
-                    columns = [column[0] for column in self._sqlite_cur.description]
+                    columns = [
+                        safe_identifier_length(column[0])
+                        for column in self._sqlite_cur.description
+                    ]
                     sql = """
                         INSERT IGNORE
                         INTO `{table}` ({fields})
                         VALUES ({placeholders})
                     """.format(
-                        table=table["name"],
+                        table=safe_identifier_length(table["name"]),
                         fields=("`{}`, " * len(columns)).rstrip(" ,").format(*columns),
                         placeholders=("%s, " * len(columns)).rstrip(" ,"),
                     )
@@ -655,7 +663,7 @@ class SQLite3toMySQL:
                     except mysql.connector.Error as err:
                         self._logger.error(
                             "MySQL transfer failed inserting data into table %s: %s",
-                            table["name"],
+                            safe_identifier_length(table["name"]),
                             err,
                         )
                         raise
